/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

#include "condor_common.h"
#include "dap_constants.h"
#include "dap_logger.h"
#include "dap_error.h"
#include "dap_server_interface.h"
#include "dap_stream_list.h"
#include "condor_netdb.h"

// This file defines functions for creating and writing logs for both the Stork
// server and individual Stork jobs.

extern char *clientagenthost;

// XXX This was deprecated the moment it was added. Don't rely on this.
// First and last entry in the global userlog list.
userlog_t *head_ulog, *tail_ulog;

// TODO: Replace me or at least let me take variadic arguments.
void write_dap_log(const char *logfilename, const char *status, const char *param1, const char *value1, const char *param2, const char *value2, const char *param3, const char *value3, const char *param4, const char *value4, const char *param5, const char *value5, const char *param6, const char *value6)
{
  
	if (! param_boolean("STORK_HISTORY_LOGGING", true) ) return;

  // -------- ClassAd stuff ------------------------------
  
  //create ClassAd
  classad::ClassAd *classad = new classad::ClassAd;
  classad::ClassAdParser parser;
  classad::ExprTree *expr = NULL;

  //insert attributes to ClassAd
  classad->Insert("timestamp",classad::Literal::MakeAbsTime());

  if ( !parser.ParseExpression(value1, expr) )
    dprintf(D_ALWAYS,"Parse error\n");
  classad->Insert(param1, expr);
  
  if ((param2 != NULL) && (value2 != NULL)){
    if ( !parser.ParseExpression(value2, expr) )
      dprintf(D_ALWAYS,"Parse error\n");
    classad->Insert(param2, expr);
  }

  if ((param3 != NULL) && (value3 != NULL)){
    if ( !parser.ParseExpression(value3, expr) )
      dprintf(D_ALWAYS,"Parse error\n");
    classad->Insert(param3, expr);
  }

  if ((param4 != NULL) && (value4 != NULL)){
    if ( !parser.ParseExpression(value4, expr) )
      dprintf(D_ALWAYS,"Parse error\n");
    classad->Insert(param4, expr);
  }

  if ((param5 != NULL) && (value5 != NULL)){
    if ( !parser.ParseExpression(value5, expr) )
      dprintf(D_ALWAYS,"Parse error\n");
    classad->Insert(param5, expr);
  }

  if ((param6 != NULL) && (value6 != NULL)){
    if ( !parser.ParseExpression(value6, expr) )
      dprintf(D_ALWAYS,"Parse error\n");
    classad->Insert(param6, expr);
  }

  if ( !parser.ParseExpression(status, expr) )
    dprintf(D_ALWAYS,"Parse error\n");
  classad->Insert("status", expr);


  //parse the ClassAd
  classad::PrettyPrint unparser;
  std::string adbuffer = "";
  unparser.Unparse(adbuffer, classad);

  //destroy the ClassAd
  //  delete expr;
  delete classad;

  // -------- End of ClassAd stuff ------------------------------

  FILE *flog;

  //write the classad to classad file
  //if ((flog = safe_fopen_wrapper(logfilename,"a+")) == NULL){
  if ((flog = fopen(logfilename,"a+")) == NULL){

   	dprintf(D_ALWAYS,
	    "cannot open logfile :%s...\n",logfilename);
    	exit(1);
  }
  fprintf (flog,"%s",adbuffer.c_str());
  fclose(flog);

}

void write_classad_log(const char *logfilename, const char *status, const classad::ClassAd *classad_)
{

  classad::ClassAd *classad = new classad::ClassAd(*classad_);
  classad::ClassAdParser parser;
  classad::ExprTree *expr = NULL;
  classad::PrettyPrint unparser;
  std::string adbuffer = "";
  FILE *flog;

  //insert attributes to ClassAd
  classad->Insert("timestamp",classad::Literal::MakeAbsTime());

  if ( !parser.ParseExpression(status, expr) )
    dprintf(D_ALWAYS,"Parse error\n");
  classad->Insert("status", expr);

  unparser.Unparse(adbuffer, classad);
  
  //write the classad to classad file
  //if ((flog = safe_fopen_wrapper(logfilename,"a+")) == NULL){
  if ((flog = fopen(logfilename,"a+")) == NULL){
      dprintf(D_ALWAYS,
	    "cannot open logfile :%s...\n",logfilename);
      exit(1);
  }

  fprintf (flog,"%s",adbuffer.c_str());

  fclose(flog);
  delete classad;
}

void write_collection_log(classad::ClassAdCollection *dapcollection, const char *dap_id, 
			  const char *update)
{
  //log the new status of the request
  classad::ClassAdParser parser;
  bool status;
  std::string key;
  key = "key = ";
  key += dap_id;
  
  std::string partial_s = "[ ";
  partial_s += update;
  partial_s += " ]";
  
  classad::ClassAd* partialAd = parser.ParseClassAd(partial_s, true);
  if (partialAd == NULL) {
	dprintf(D_ALWAYS, "ParseClassAd partial %s failed! dap_id:%s\n",
		partial_s.c_str(), dap_id);
	return;
  }
  status = partialAd->Insert("timestamp",classad::Literal::MakeAbsTime());
  if (status == false) {
	  	delete partialAd;
		dprintf(D_ALWAYS,
			"ParseClassAd partial %s insert timestamp failed! dap_id:%s\n",
			partial_s.c_str(), dap_id);
	  return;
  }
    
  classad::PrettyPrint unparser;
  std::string adbuffer = "";
  unparser.Unparse(adbuffer, partialAd);

  std::string modify_s ="[ Updates  = ";
  modify_s += adbuffer;
  modify_s += " ]";
  
  classad::ClassAd* tmpad = parser.ParseClassAd(modify_s, true);
  if (tmpad == NULL) {
	  delete partialAd;
	  dprintf(D_ALWAYS, "ParseClassAd modify %s failed! dap_id:%s\n",
			partial_s.c_str(), dap_id);
	  return;
  }

  unparser.Unparse(adbuffer, tmpad);
		  

  if (!dapcollection->ModifyClassAd(key, tmpad)){ // deletes ModifyClassAd()
    dprintf(D_ALWAYS, "ModifyClassAd failed! dap_id:%s\n", dap_id);
  }

  delete partialAd;
  //delete tmpad;	// delete'ed by ModifyClassAd()
  return;
}

// Userlog writing functions
// -------------------------
// These functions are responsible for writing userlog entries. The ulog family
// of functions here takes a pointer to string pointer and a job ad. They
// dynamically allocate a string containing the body of the userlog entry and
// point b_str to it. They return true on success, false otherwise.

// When a job has just been submitted...
bool ulog_submit(char **b_str, const classad::ClassAd *ad) {
	int ret;
	std::string host, notes;
	ad->EvaluateAttrString("submit_host", host);
	ad->EvaluateAttrString("LogNotes", notes);
	ret = notes.empty() ?
	      asprintf(b_str, "Job submitted from host: %s", host.c_str()) :
	      asprintf(b_str, "Job submitted from host: %s\n    %.8191s",
	               host.c_str(), notes.c_str());
	return ret != -1;
}

// When a job has just begun executing...
bool ulog_execute(char **b_str, const classad::ClassAd *ad) {
	std::string host;
	ad->EvaluateAttrString("execute_host", host);
	return asprintf(b_str, "Job executing on host: %s", host.c_str()) != -1;
}

// When a job has just terminated. The userlog entry body should also contain
// the number of bytes transferred, but we don't support that yet.
bool ulog_term(char **b_str, const classad::ClassAd *ad) {
	std::string tt;
	bool normal = false;
	int exit_status, retval;

	ad->EvaluateAttrString("termination_type", tt);

	// Act depending on termination_type
	if (tt == "job_retry_limit" || tt == "server_error") {
		normal = false;
		retval = EXIT_FAILURE;
	} else if (tt == "exited") {
		// The job exited. Parse the exit status.
		ad->EvaluateAttrInt("exit_status", exit_status);
		if (WIFEXITED(exit_status)) {
			retval = WEXITSTATUS(exit_status);
			normal = (retval == EXIT_SUCCESS);
		} else if (WIFSIGNALED(exit_status)) {
			retval = WTERMSIG(exit_status);
			normal = false;
		} else return false;  // Unknown exit status
	} else return false;  // Unknown termination type

	return asprintf(b_str, (normal) ?
	                "Normal termination (return value %d)" :
	                "Abnormal termination (signal %d)",
	                retval) != -1;
}

// When a job has been aborted...
bool ulog_abort(char **b_str, const classad::ClassAd*) {
	*b_str = strdup("Job was aborted by the user.");
	return *b_str != NULL;
}

// When we'd like to record anything else in the userlog for any reason...
bool ulog_generic(char **b_str, const classad::ClassAd *ad) {
	std::string generic_event;
	ad->EvaluateAttrString("generic_event", generic_event);

	*b_str = strdup(generic_event.c_str());
	return *b_str != NULL;
}

// Record an entry in the specified userlog. A userlog entry follows this
// format:
//   HEADER BODY
//   ...
//
// Where the format of HEADER is:
//   event (dap_id.proc.subproc) mm/dd hh:mm:ss
//
// And BODY is determined based on the event. The ... is called the event
// delimiter.
bool write_ulog_event(userlog_t *ulog, int event, const classad::ClassAd *ad) {
	time_t timer;
	struct tm tm;
	int ret;
	char *h_str = NULL, *b_str = NULL;  // Head and body strings

	// Fill in the time struct with the current date and time
	if (time(&timer) == -1) return false;
	tm = *localtime(&timer);
	if (mktime(&tm) == -1) return false;

	// Write userlog header to intermediate buffer
	asprintf(&h_str, "%03d (%03d.%03d.%03d) %02d/%02d %02d:%02d:%02d",
	         event, ulog->dap_id, ulog->proc, 0,
	         tm.tm_mon+1, tm.tm_mday,
	         tm.tm_hour, tm.tm_min, tm.tm_sec);

	// Write userlog body to intermediate buffer
	switch (event) {
		case ULOG_SUBMIT:  ret = ulog_submit(&b_str, ad); break;
		case ULOG_EXECUTE: ret = ulog_execute(&b_str, ad); break;
		case ULOG_TERM:    ret = ulog_term(&b_str, ad); break;
		case ULOG_ABORT:   ret = ulog_abort(&b_str, ad); break;
		case ULOG_GENERIC: ret = ulog_generic(&b_str, ad); break;
		default:           ret = false; goto fail;
	}

	// Check if there was an error
	if (!ret) goto fail;

	// Broadcast entry to all registered streams
	sl_printf(ulog->out, "%s %s\n...\n", h_str, b_str);

	// Clean up and exit
	free(h_str);
	free(b_str);
	return true;

	// Error handling
	fail:
	dprintf(D_ALWAYS, "ERROR writing userlog entry for job %d, event %d\n",
	        ulog->dap_id, event);
	free(h_str);
	free(b_str);
	return false;
}

// Userlog state functions
// -----------------------
// Functions and structures for maintaining userlog state information for Stork
// jobs.

// new_ulog will return a dynamically allocated userlog_t with its members
// initialized according to the passed ad. The returned userlog_t should have
// close_ulog called on it after it is no longer needed. Returns a fully
// initialized userlog_t if no error occurred. Otherwise returns NULL.
//
// It will also append it to the global list of userlogs. It won't do this soon.
userlog_t *new_ulog(const classad::ClassAd *ad) {
	userlog_t *ulog;
	std::string owner, ulog_path;

	// Allocate some (zeroed) space for our struct
	ulog = (userlog_t*) calloc(1, sizeof(userlog_t));
	if (!ulog) goto fail;

	// Get owner name. This is dynamically allocated!
	if (ad->EvaluateAttrString("owner", owner) && !owner.empty())
		ulog->owner = strdup(owner.c_str());  // Free me!
	else goto fail;

	// Initialize stream list
	ulog->out = new_stream_list();
	if (!ulog->out) {
		dprintf(D_ALWAYS, "new_ulog: Could not init stream list!\n");
		goto fail;
	}
	
	// Get job identifiers
	ad->EvaluateAttrInt("dap_id", ulog->dap_id);
	ad->EvaluateAttrInt("pid", ulog->proc);
	ad->EvaluateAttrBool("log_xml", ulog->log_xml);

	return ulog;

	fail:
	if (ulog) free(ulog->owner);
	free(ulog);
	return NULL;
}

// Free a pointer to a userlog and close/free its resources.
void close_ulog(userlog_t *ulog) {
	if (!ulog) return;
	free(ulog->owner);
	if (ulog->out) sl_close(ulog->out);
	free(ulog);
}

// Userlog list functions
// ----------------------
// These are functions necessary for maintaining the global userlog list.
//
// It's a dumb temporary workaround until we settle on a way of storing job
// state information that isn't conveniently a string we can put in a ClassAd.
// It uses a dumb O(n) search to find the relevant userlog for a given dap_id,
// because it won't be here next release and no one will even notice.

// Add a userlog_t to end of the global list
void ulog_list_add(userlog_t *ul) {
	if (!ul) return;

	if (tail_ulog) {  // Something in the list
		ul->prev = tail_ulog;
		tail_ulog->next = ul;
		tail_ulog = ul;
	} else head_ulog = tail_ulog = ul;
}

// Remove and destroy a userlog_t in the global list
void ulog_list_del(userlog_t *ul) {
	if (!ul) return;
	ulog_unlink(ul);
	close_ulog(ul);
}

// And do it by its dap_id, too
void ulog_list_del(int dap_id) {
	ulog_list_del(ulog_list_find(dap_id));
}

// Link a userlog_t from its neighbors and zero its prev and next
void ulog_unlink(userlog_t *ul) {
	if (!ul) return;
	if (ul == head_ulog) head_ulog = ul->next;
	if (ul == tail_ulog) tail_ulog = ul->prev;
	if (ul->prev) ul->prev->next = ul->next;
	if (ul->next) ul->next->prev = ul->prev;
	ul->prev = ul->next = NULL;
}

// Search the userlog list for a userlog_t with a give dap_id. Returns NULL if
// nothing can be found. Stupid O(n) search because we don't even care.
userlog_t *ulog_list_find(int dap_id) {
	for (userlog_t *ul = head_ulog; ul; ul = ul->next)
		if (ul->dap_id == dap_id) return ul;
	return NULL;
}

// Userlog interface functions
// ---------------------------
// These are the only userlog functions which should ever be used in other
// files. They allow userlogs to be registered and have files and sockets
// associated with them, and allow writing to registered userlogs.

// Create a new userlog for a job if one doesn't already exist, and register a
// stream with it (or not, if obj is NULL)
bool ulog_register(const classad::ClassAd *ad, void *obj, int type) {
	if (!ad) return false;

	int dap_id;
	userlog_t *ulog;
	ad->EvaluateAttrInt("dap_id", dap_id);

	// Find or create a userlog
	if (ulog = ulog_list_find(dap_id));
	else if (ulog = new_ulog(ad))
		ulog_list_add(ulog);
	else return false;

	return obj ? sl_add(ulog->out, obj, type) : true;
}

bool ulog_register(const classad::ClassAd *ad, FILE *fp) {
	return ulog_register(ad, fp, SL_FILE_TYPE);
}

bool ulog_register(const classad::ClassAd *ad, ReliSock *s) {
	return ulog_register(ad, s, SL_SOCK_TYPE);
}

// Allow a path to be passed, and open a file
bool ulog_register(const classad::ClassAd *ad, const char *path) {
	FILE *fp;

	fp = fopen(path, "a");
	if (!fp) return false;

	return ulog_register(ad, fp, SL_FILE_TYPE);
}

// Create a new userlog for a job and register its log attribute if it's there
// TODO: stork_submit should ideally strip the log attribute from the submit
// ad iff it's submitting to a remote host. Otherwise, it should be present to
// indicate that a log should be kept locally. We should first check if a job is
// local, and if so check for the log attribute, and register it if it's there.
bool ulog_register(const classad::ClassAd *ad) {
	if (!ad) return false;

	std::string log;

	// Check for "log" attribute
	//if (ad->EvaluateAttrString("log", log) && !log.empty())
	//	return ulog_register(ad, log.c_str());
	return ulog_register(ad, NULL, -1);
}

// Unregister a userlog and close all of the associated files and sockets
void ulog_unregister(int dap_id) {
	ulog_list_del(dap_id);
}

void ulog_unregister(const classad::ClassAd *ad) {
	if (!ad) return;

	int dap_id;
	ad->EvaluateAttrInt("dap_id", dap_id);
	
	ulog_unregister(dap_id);
}

// This function is called by Stork to write a userlog entry to all streams
// registered to receive userlog information for the job.
bool userlog(const classad::ClassAd *ad, int event) {
	int ret, dap_id;
	userlog_t *ulog;
	
	// Find a userlog for the job
	ad->EvaluateAttrInt("dap_id", dap_id);
	ulog = ulog_list_find(dap_id);
	if (!ulog) return false;

	// Make sure there's something to write to
	if (!ulog->out) return true;  // Nothing to do!
	
	// Write the event to the userlog
	ret = write_ulog_event(ulog, event, ad);
	if (!ret)
		dprintf(D_ALWAYS, "userlog: could not write to userlog\n");
	return ret;
}
