/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

#ifndef _DAP_STREAM_LIST_H
#define _DAP_STREAM_LIST_H

#include "condor_common.h"
#include "condor_debug.h"
#include "dap_error.h"
#include "sock.h"
#include "reli_sock.h"

// The two types of objects which can be stored in this list
#define SL_FILE_TYPE 1
#define SL_SOCK_TYPE 2

struct stream_node {
	int type;
	void *obj;
	stream_node *prev, *next;
};

struct stream_list {
	char *tmplog;
	stream_node *head, *tail;
};

stream_node *new_stream_node(void *obj, int type);
stream_list *new_stream_list();

void sl_close(stream_list *sl);

bool sl_add(stream_list *sl, stream_node *sn);
bool sl_add(stream_list *sl, void *obj, int type);
bool sl_add(stream_list *sl, ReliSock *sock);
bool sl_add(stream_list *sl, FILE *fp);

void print_list(stream_list *sl);

void sl_del(stream_list *sl, void *obj);

void sl_printf(stream_list *sl, char *fmt, ...);

void sl_put(stream_list *sl, char *str);

void sl_write(stream_list *sl, void *bytes, int len);

void sl_node_write(stream_node *sn, void *bytes, int len);

#endif
