/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include <sys/socket.h>
#include "dap_constants.h"
#include "dap_utility.h"
#include "my_hostname.h"

// This file includes a bunch of helpful utility functions to make doing
// tedious and common tasks easy.

// FIXME: This whole thing is kinda bad.

// Strip extra quotes and spaces from the front and back of the string. Yes,
// this modifies the string in place... Return str again.
char *strip_str(char *str) {
	char *bad = " \t\"'";  // Remove these characters
	char *str_a, *str_z;  // Pointers to start and end of str.

	// Basic input checking (no NULLs or empty strings allowed!)
	if (!str || !*str) return str;

	// Start at the beginning of the string, and increment the str_a pointer
	// until the first good character (or a null character) is found.
	for (str_a = str; *str_a && strchr(bad, *str_a); str_a++);
	if (!*str_a) return &(*str = '\0');  // If empty, set str = "" and return

	// If we get here, there's more left in the string. Starting at the end,
	// decrement the str_z pointer until we reach the new location for the
	// terminating null.
	for (str_z = strchr(str_a, '\0'); strchr(bad, str_z[-1]); str_z--);
	*str_z = '\0';

	// Now copy the stripped string back into str if it's necessary
	if (str < str_a) memmove(str, str_a, strlen(str_a)+1);

	return str;
}

// Parse a URL out into its individual components. If any of the arguments are
// NULL, then simply ignore them.
// FIXME: This needs to do some kind of bounds checking or return dymanically
// allocated strings. Or, better yet, create a Url object or struct.
void parse_url(const char *url, char *protocol, char *host, char *filename) {
	char *temp_url = NULL;
	char *_protocol;  // Even if protocol is NULL, we still need to know it
	char *p;  // A general purpose pointer to parts of temp_url

	// Input checking
	if (!url || !*url) goto fail;  // NULL/empty string check

	// Initialize
	if (protocol) *protocol = '\0';
	if (host) *host = '\0';
	if (filename) *filename = '\0';

	// Make a copy of the url to operate on, and strip it
	temp_url = strip_str(strdup(url));  // Free me!
	if (!temp_url || !*temp_url) goto fail;

	// Get protocol
	if (p = strtok(temp_url, ":"))
		_protocol = p;  // Save for later
	else
		goto fail;
	if (protocol) strcpy(protocol, p);

	// Get host
	if (!strcmp(_protocol, "file")) { // Special case for file
		strcpy(host, "localhost");
	} else if (p = strtok(NULL, "/")) {
		if (host) strcpy(host, p);
	} else {
		goto fail;
	}
	
	// Get filename. Prepend / if there's not one.
	if (filename && (p = strtok(NULL, ""))) {
		if (*p != '/')
			sprintf(filename, "/%s", p);
		else
			strcpy(filename, p);
	}

	// Clean up and exit
	free(temp_url);
	return;

	// Go here to clean up before failing
	fail:
	free(temp_url);
	printf("Error in parsing URL \"%s\"\n", url);
}

// Create a predictable unique path, given a directory, basename, job id, and
// pid. The return value points to a statically allocated string. This
// function is not reentrant.
const char *job_filepath(const char *basename, const char *suffix,
                         const char *dap_id, pid_t pid) {
	static char path[_POSIX_PATH_MAX];

	sprintf(path, "%s%s-job%s-pid%u",
	        basename, suffix, dap_id, pid);
	return path;
}

// Exactly like the interface to send(), though this function will ensure that
// the entire length is written through multiple calls to send(). Unless an
// error occurred, this should always return len.
int sendall(int s, char *buf, int len, int flags) {
	int ret, sent = 0;

	do {
		ret = send(s, &buf[sent], len - sent, flags);
		if (ret == -1) return sent;
		sent += ret;
	} while (sent < len);

	return len;
}
